#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2024-07-08
# @Filename: mwm_snc_further_ext.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


class MWM_snc_further_ext_Base_Carton(BaseCarton):
    """
    MWM_snc_further_ext_Base_Carton is a base carton.

    Actual cartons are implemented as subclasses of MWM_SNC_further_ext_Base_Carton
    for different conditions in the WHERE clause.

    https://sdss-wiki.atlassian.net/wiki/spaces/MWM/pages/132579391/A+Further+Extension+of+the+SNC
    """

    def build_query(self, version_id, query_region=None):
        ll = Gaia_DR3.l
        bb = Gaia_DR3.b

        # Dense regions (Galactic plane, SMC, LMC).
        gal_cut = (
            ((ll <= 180) & (bb < (-0.139 * ll + 25)) & (bb > (0.139 * ll - 25)))
            | ((ll > 180) & (bb > (-0.139 * ll + 25)) & (bb < (0.139 * ll - 25)))
            | (fn.sqrt(fn.pow(ll - 303.2, 2) + 2 * fn.pow(bb + 44.4, 2)) < 5)
            | (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
        )

        cte = (
            Gaia_DR3.select(Gaia_DR3.source_id)
            .where(
                Gaia_DR3.parallax < 10,
                Gaia_DR3.parallax - Gaia_DR3.parallax_error > 2,
                Gaia_DR3.phot_g_mean_mag + 5 * fn.log10(Gaia_DR3.parallax / 1000) + 5
                > 19.125 - 0.03225 * (1000 / Gaia_DR3.parallax),
            )
            .cte("plx_mag_cte", materialized=True)
        )

        query = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.l,
                Gaia_DR3.b,
                Gaia_DR3.parallax,
                Gaia_DR3.parallax_error,
                Gaia_DR3.phot_g_mean_mag.alias("gaia_dr3_phot_g_mean_mag"),
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag,
                Gaia_DR3.ruwe,
                Gaia_DR3.phot_bp_rp_excess_factor,
                Gaia_DR3.astrometric_excess_noise,
            )
            .join(CatalogToGaia_DR3)
            .join_from(Gaia_DR3, cte, on=(Gaia_DR3.source_id == cte.c.source_id))
            .where(
                (
                    (Gaia_DR3.astrometric_excess_noise < 2)
                    & (Gaia_DR3.ruwe < 1.2)
                    & (
                        Gaia_DR3.phot_bp_rp_excess_factor
                        > 1
                        + 0.015 * fn.pow(Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag, 2)
                    )
                    & (
                        Gaia_DR3.phot_bp_rp_excess_factor
                        < 1.3
                        + 0.06 * fn.pow(Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag, 2)
                    )
                    & gal_cut
                )
                | ~gal_cut
            )
            .where(CatalogToGaia_DR3.version_id == version_id, CatalogToGaia_DR3.best >> True)
            .with_cte(cte)
        )
        return query


class Openfibertargets_mwm_snc_further_ext_apogee_bright_Carton(MWM_snc_further_ext_Base_Carton):
    """Shorthand name: openfibertargets_mwm_snc_further_ext_apogee_bright
    REPLACES openfibertargets_nov2020_24
    Return columns: Gaia Galactic longitude and latitude, parallax, parallax_error,
        G, BP, RP, ruwe, phot_bp_rp_excess_factor, astrometric_excess_noise, and 2MASS H
    Metadata: can_offset=True
    Priority: 6085
    Cadence: bright_1x1
    Instrument: APOGEE
    Program: open_fiber
    Lead contact: Ilija Medan
    """

    name = "openfibertargets_mwm_snc_further_ext_apogee_bright"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    program = "open_fiber"
    mapper = "MWM"
    priority = 6085
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = (
            query.switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC)
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < 11,
            )
        )

        return query


class Openfibertargets_mwm_snc_further_ext_apogee_dark_Carton(MWM_snc_further_ext_Base_Carton):
    """Shorthand name: openfibertargets_mwm_snc_further_ext_apogee_dark
    REPLACES openfibertargets_nov2020_24
    Return columns: Gaia Galactic longitude and latitude, parallax, parallax_error,
        G, BP, RP, ruwe, phot_bp_rp_excess_factor, astrometric_excess_noise, and 2MASS H
    Metadata: can_offset=True
    Priority: 6085
    Cadence: dark_1x1
    Instrument: APOGEE
    Program: open_fiber
    Lead contact: Ilija Medan
    """

    name = "openfibertargets_mwm_snc_further_ext_apogee_dark"
    category = "science"
    instrument = "APOGEE"
    cadence = "dark_1x1"
    program = "open_fiber"
    mapper = "MWM"
    priority = 6085
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = (
            query.switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC)
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < 11,
            )
        )

        return query


class Openfibertargets_mwm_snc_further_ext_boss_Carton(MWM_snc_further_ext_Base_Carton):
    """Shorthand name: openfibertargets_mwm_snc_further_ext_boss
    REPLACES openfibertargets_nov2020_24
    Return columns: Gaia Galactic longitude and latitude, parallax, parallax_error,
    G, BP, RP, ruwe, phot_bp_rp_excess_factor, astrometric_excess_noise, and 2MASS H
    Metadata: can_offset=True
    Priority: 6086
    Cadence: bright_1x1 for stars with G < 16 and dark_1x1 for stars with G > 16
    Instrument: BOSS
    Program: open_fiber
    Lead contact: Ilija Medan
    """

    name = "openfibertargets_mwm_snc_further_ext_boss"
    category = "science"
    instrument = "BOSS"
    cadence = None  # cadence is set in post_process()
    program = "open_fiber"
    mapper = "MWM"
    priority = 6086
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)

        query = query.where(
            Gaia_DR3.phot_g_mean_mag > 10,
            Gaia_DR3.phot_g_mean_mag < 20,
            Gaia_DR3.radial_velocity.is_null(),
        )

        return query

    def post_process(self, model):
        # Note that self is a parameter of post_process() above but
        # self is not a parameter of post_process() below
        # since we are using super().
        super().post_process(model)

        # set cadence bright_1x1 for stars with G < 16 and dark_1x1 for stars with G > 16

        cursor = self.database.execute_sql(
            "select catalogid, gaia_dr3_phot_g_mean_mag from "
            + " sandbox.temp_openfibertargets_mwm_snc_further_ext_boss ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            current_g = output[i][1]

            if current_g < 16:
                current_cadence = "bright_1x1"
            else:
                current_cadence = "dark_1x1"

            if current_cadence is not None:
                self.database.execute_sql(
                    " update sandbox.temp_openfibertargets_mwm_snc_further_ext_boss "
                    + " set cadence = '"
                    + current_cadence
                    + "'"
                    " where catalogid = " + str(current_catalogid) + ";"
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-31
# @Filename: mwm_ob.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTIC_v8,
    CatalogToTwoMassPSC,
    GAIA_ASSAS_SN_Cepheids,
    Gaia_DR2,
    Gaia_DR2_TwoMass_Best_Neighbour,
    Gaia_DR3,
    Gaia_dr3_astrophysical_parameters,
    TIC_v8,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


TMBN = Gaia_DR2_TwoMass_Best_Neighbour


class MWM_OB_Core_Base_Carton(BaseCarton):
    """
    This base carton contains the part of the below query.
    which is common to all the mmw_ob_core cartons

    The derived cartons are as below.

    mwm_ob_core_boss
    Teff_esphs > = 10000
    CADENCE: bright_3x1
    PRIORITY: 1800
    INSTRUMENT:BOSS
    can_offset=True

    mwm_ob_core_boss_single
    Teff_esphs > = 10000
    CADENCE: bright_1x1
    PRIORITY: 1801
    INSTRUMENT:BOSS
    can_offset=True

    mwm_astar_core_boss
    (Teff_esphs < 10000) or (Teff_esphs is NULL)
    CADENCE: bright_3x1
    PRIORITY: 2800
    INSTRUMENT:BOSS
    can_offset=True

    mwm_astar_core_boss_single
    (Teff_esphs < 10000) or (Teff_esphs is NULL)
    CADENCE: bright_1x1
    PRIORITY: 2801
    INSTRUMENT:BOSS
    can_offset=True

    Milky Way OB stars.

    Definition: Select all the hot, young stars in Gaia and 2MASS  with M_K < 0
    mag (M ~ 4 M_Sun) with Gaia G < 16 mag in the Milky Way, then subsampled
    to reduce the sample size / observing burden.

    Query on the Gaia archive (this is the v0.5 query, v1 is slightly modified):

        SELECT  g.source_id
        FROM gaiadr3.gaia_source as g
        JOIN gaiadr3.ruwe AS r
            USING (source_id)
        INNER JOIN gaiadr3.tmass_best_neighbour AS xmatch
            ON g.source_id = xmatch.source_id
        INNER JOIN gaiadr1.tmass_original_valid AS tm
            ON tm.tmass_oid = xmatch.tmass_oid
        WHERE parallax < fn.pow(10, ((10. - tm.ks_m - 0.61) / 5.))
            AND g.phot_g_mean_mag < 16.
            AND tm.j_m - tm.ks_m - 0.25 * (g.phot_g_mean_mag - tm.ks_m) < 0.10
            AND tm.j_m - tm.ks_m - 0.25 * (g.phot_g_mean_mag - tm.ks_m) > -0.30
            AND tm.j_m - tm.h_m < 0.15 * (g.phot_g_mean_mag - tm.ks_m) + 0.05
            AND tm.j_m - tm.h_m > 0.15 * (g.phot_g_mean_mag - tm.ks_m) - 0.15
            AND tm.j_m - tm.ks_m < 0.23 * (g.phot_g_mean_mag - tm.ks_m) + 0.03
            AND g.phot_g_mean_mag > 2 * (g.phot_g_mean_mag - tm.ks_m) + 3.0

    Notes:
        - ks_m is the same as twomass_psc.h_m.

    """

    def build_query(self, version_id, query_region=None):
        km = TwoMassPSC.k_m
        hm = TwoMassPSC.h_m
        jm = TwoMassPSC.j_m
        Gm = Gaia_DR3.phot_g_mean_mag

        query = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                km,
                hm,
                jm,
                Gm,
                Gaia_dr3_astrophysical_parameters.teff_esphs,
                Gaia_DR3.parallax,
                Gaia_DR3.source_id,
                km.alias("ks_m"),
            )
            .join(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToTwoMassPSC.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(TwoMassPSC)
            .join_from(
                Gaia_DR3,
                Gaia_dr3_astrophysical_parameters,
                on=(Gaia_dr3_astrophysical_parameters.source_id == Gaia_DR3.source_id),
            )
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
            )
            .where(
                Gaia_DR3.parallax < fn.pow(10, ((10.0 - km - 0.61) / 5.0)),
                Gm < 16.0,
                jm - km - 0.25 * (Gm - km) < 0.10,
                jm - km - 0.25 * (Gm - km) > -0.30,
                jm - hm < 0.15 * (Gm - km) + 0.05,
                jm - hm > 0.15 * (Gm - km) - 0.15,
                jm - km < 0.23 * (Gm - km) + 0.03,
                Gm > 2 * (Gm - km) + 3.0,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_OB_Core_boss_Carton(MWM_OB_Core_Base_Carton):
    """
    mwm_ob_core_boss
    Teff_esphs > = 10000
    CADENCE: bright_3x1
    PRIORITY: 1800
    INSTRUMENT:BOSS
    can_offset=True
    """

    name = "mwm_ob_core_boss"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_3x1"
    program = "mwm_ob"
    priority = 1800
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(Gaia_dr3_astrophysical_parameters.teff_esphs >= 10000)
        return query


class MWM_OB_Core_boss_single_Carton(MWM_OB_Core_Base_Carton):
    """
    mwm_ob_core_boss_single
    Teff_esphs > = 10000
    CADENCE: bright_1x1
    PRIORITY: 1801
    INSTRUMENT:BOSS
    can_offset=True
    """

    name = "mwm_ob_core_boss_single"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_1x1"
    program = "mwm_ob"
    priority = 1801
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(Gaia_dr3_astrophysical_parameters.teff_esphs >= 10000)
        return query


class MWM_astar_Core_boss_Carton(MWM_OB_Core_Base_Carton):
    """
    mwm_astar_core_boss
    (Teff_esphs < 10000) or (Teff_esphs is NULL)
    CADENCE: bright_3x1
    PRIORITY: 2800
    INSTRUMENT:BOSS
    can_offset=True
    """

    name = "mwm_astar_core_boss"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_3x1"
    program = "mwm_ob"
    priority = 2800
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            (Gaia_dr3_astrophysical_parameters.teff_esphs < 10000)
            | (Gaia_dr3_astrophysical_parameters.teff_esphs.is_null())
        )
        return query


class MWM_astar_Core_boss_single_Carton(MWM_OB_Core_Base_Carton):
    """
    mwm_astar_core_boss_single
    (Teff_esphs < 10000) or (Teff_esphs is NULL)
    CADENCE: bright_1x1
    PRIORITY: 2801
    INSTRUMENT:BOSS
    can_offset=True
    """

    name = "mwm_astar_core_boss_single"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_1x1"
    program = "mwm_ob"
    priority = 2801
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            (Gaia_dr3_astrophysical_parameters.teff_esphs < 10000)
            | (Gaia_dr3_astrophysical_parameters.teff_esphs.is_null())
        )
        return query


class MWM_OB_Cepheids_boss_Carton(BaseCarton):
    """Milky Way Cepheids.

    Definition: List of Cepheids compiled by Inno et al. (in prep). The
    catalogue is obtained by using Gaia DR2 and ASAS-SN. Gaia parallaxes and
    variability information are used to select an initial sample, for which
    multi-epoch V-band photometry from the ASAS-SN survey is retrieved. From
    the ASAS-SN lightcurves, periods and Fourier parameters are derived. The
    Fourier parameters are used to identify classical Cepheids. The final
    catalogue is assembled by including several public classic Cepheid
    databases: OGLE-cCs, Gaia-cCs, ASASSN-cCs, VSX-cCs, Simbad-cCs,  and
    WISE-cCs. The final catalogue consists of ~3000 targets, between
    6 < G < 17 mag.

    """

    name = "mwm_ob_cepheids_boss"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_3x1"
    program = "mwm_ob"
    priority = 1800
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            GAIA_ASSAS_SN_Cepheids.select(CatalogToTIC_v8.catalogid, Gaia_DR2.source_id)
            .join(Gaia_DR2)
            .join(TIC_v8)
            .join(CatalogToTIC_v8)
            .where(CatalogToTIC_v8.version_id == version_id, CatalogToTIC_v8.best >> True)
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-04-25
# @Filename: mwm_bin_vis.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import JOIN

from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    Galah_dr3,
    Lamost_dr6,
    RAVE_DR6_Gaia_DR3_XMatch,
    RAVE_DR6_Gauguin_Madera,
    SDSS_DR17_APOGEE_Allstarmerge,
    TwoMassPSC,
    Visual_binary_gaia_dr3,
)

from target_selection.cartons import BaseCarton


class MWM_Wide_Binaries_Base_Carton(BaseCarton):
    """
    This is the MWM wide binaries base carton.
    The derived cartons are
    mwm_bin_vis_apogee
    mwm_bin_vis_boss
    """

    def build_query(self, version_id, query_region=None):
        # This carton is a bit complicated because each entry in visual_binary_gaia_dr3 is
        # actually two targets (source_id1, source_id2) that may need to be observed.
        # In the main query what we do is, for each one of those sources, collect all the
        # information about whether they have been observed with APOGEE, LAMOST, etc. while
        # keeping the visual_binary_gaia_dr3 information.

        query = None

        for col in ["source_id1", "source_id2"]:
            source_col = getattr(Visual_binary_gaia_dr3, col)
            this_query = (
                CatalogToGaia_DR3.select(
                    CatalogToGaia_DR3.catalogid,
                    Gaia_DR3.source_id,
                    Visual_binary_gaia_dr3.source_id1,
                    Visual_binary_gaia_dr3.source_id2,
                    TwoMassPSC.h_m,
                    SDSS_DR17_APOGEE_Allstarmerge.apogee_id,
                    Galah_dr3.dr3_source_id.alias("galah_dr3_source_id"),
                    Lamost_dr6.source_id.alias("lamost_dr6_source_id"),
                    RAVE_DR6_Gauguin_Madera.rave_obs_id,
                    Visual_binary_gaia_dr3.dr2_parallax1,
                    Visual_binary_gaia_dr3.dr2_parallax2,
                    Visual_binary_gaia_dr3.sep_au,
                )
                .join(Gaia_DR3)
                .join(Visual_binary_gaia_dr3, on=(Gaia_DR3.source_id == source_col))
                .join_from(
                    CatalogToGaia_DR3,
                    CatalogToTwoMassPSC,
                    on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
                )
                .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
                .join(
                    SDSS_DR17_APOGEE_Allstarmerge,
                    on=(
                        TwoMassPSC.designation
                        == peewee.fn.replace(SDSS_DR17_APOGEE_Allstarmerge.apogee_id, "2M", "")
                    ),
                    join_type=JOIN.LEFT_OUTER,
                )
                .join_from(
                    Visual_binary_gaia_dr3,
                    Galah_dr3,
                    on=(source_col == Galah_dr3.dr3_source_id),
                    join_type=JOIN.LEFT_OUTER,
                )
                .join_from(
                    Visual_binary_gaia_dr3,
                    Lamost_dr6,
                    on=(source_col == Lamost_dr6.source_id),
                    join_type=JOIN.LEFT_OUTER,
                )
                .join_from(
                    Visual_binary_gaia_dr3,
                    RAVE_DR6_Gaia_DR3_XMatch,
                    on=(source_col == RAVE_DR6_Gaia_DR3_XMatch.gaiae3),
                    join_type=JOIN.LEFT_OUTER,
                )
                .join(
                    RAVE_DR6_Gauguin_Madera,
                    on=(RAVE_DR6_Gaia_DR3_XMatch.obsid == RAVE_DR6_Gauguin_Madera.rave_obs_id),
                    join_type=JOIN.LEFT_OUTER,
                )
                .where(
                    CatalogToGaia_DR3.version_id == version_id,
                    CatalogToGaia_DR3.best >> True,
                    CatalogToTwoMassPSC.version_id == version_id,
                    CatalogToTwoMassPSC.best >> True,
                )
                .distinct(Gaia_DR3.source_id)
            )

            if "apogee" in self.name:
                this_query = this_query.where(TwoMassPSC.h_m <= self.parameters["h_m_limit"])
            else:
                this_query = this_query.where(TwoMassPSC.h_m > self.parameters["h_m_limit"])

            if query is None:
                query = this_query
            else:
                query |= this_query

        return query


class MWM_Wide_Binaries_APOGEE_Carton(MWM_Wide_Binaries_Base_Carton):
    """Wide binaries APOGEE carton."""

    mapper = "MWM"
    category = "science"
    program = "mwm_filler"
    cadence = "bright_1x1"
    name = "mwm_bin_vis_apogee"
    instrument = "APOGEE"
    priority = 2999


class MWM_Wide_Binaries_BOSS_Carton(MWM_Wide_Binaries_Base_Carton):
    """Wide binaries BOSS carton."""

    mapper = "MWM"
    category = "science"
    program = "mwm_filler"
    cadence = "bright_1x1"
    name = "mwm_bin_vis_boss"
    instrument = "BOSS"
    priority = 2999



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
#          Updated for v1 by dwelly@mpe.mpg.de
# @Date: 2021-01-12
# @Filename: mwm_yso.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    EROSITASupersetv1Compactobjects,
    EROSITASupersetv1Stars,
    Gaia_DR3,
)

from target_selection.cartons import BaseCarton


# from target_selection.exceptions import TargetSelectionError

# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# For example
# peewee Model name ---> postgres table name
# AllWise ---> 'catalogdb.allwise'
# EROSITASupersetAGN --->'catalogdb.erosita_superset_agn'
# EROSITASupersetClusters ---> 'catalogdb.erosita_superset_clusters'
# EROSITASupersetStars ---> 'catalogdb.erosita_superset_stars'
# EROSITASupersetCompactobjects ---> 'catalogdb.erosita_superset_compactobjects'
# Gaia_DR2 --->'catalogdb.gaia_dr2_source'
# TwoMassPSC --->'catalogdb.twomass_psc'


class MWM_EROSITA_Stars_Boss_Carton(BaseCarton):
    """MWM eROSITA Stars
    Owner: Lead by MWM (with assistance from BHM?)?? Jennifer Johnson, Tom Dwelly

    Shorthand name: mwm_erosita_stars_boss

    What is it?: Optical counterparts to eROSITA sources with
    a high likelihood to be stars. Stellar identification are
    based on Bayesian scheme using the Gaia edr3 and Tycho catalogs.
    Those objects shall be observed to increase the legacy value
    of eROSITA / SDSS; possible science questions relate to
    Halpha emission although EW are expected to be around 1 Angstrom or less.

    Simplified Description of selection criteria:
    "Select all objects with xmatch_metric>0.0 and
    xmatch_method = 'HamStar' and xmatch_version = 'v1.1.1'
    from catalogdb.erosita_superset_v1_stars that are
    in a magnitude range amenable for SDSS-V BOSS spectroscopy
    (suitability estimated from lo < Gmag < hi)"

    Wiki page: N/A

    Additional source catalogs needed: None

    Additional cross-matching needed: No

    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    boss_bright_1xN or boss_dark_1xN
    (moon phase and number of exposures depends on target magnitude).
    No detailed time spacing required.

    Pseudo SQL (optional):

    SELECT c.catalogid, c.ra, c.dec,
    gaia.phot_g_mean_mag, gaia.phot_bp_mean_mag,
    gaia.phot_rp_mean_mag, gaia.parallax, gaia.pmra, gaia.pmdec,
    estars.target_priority, estars.xmatch_metric, estars.ero_flux,
    estars.ero_ra, estars.ero_dec, estars.opt_ra, estars.opt_dec
     FROM catalog c
     INNER JOIN catalog_to_gaia_dr3_source c2g3 USING (catalogid)
     INNER JOIN gaia_dr3_source g3 ON g3.source_id = c2g3.target_id
     INNER JOIN erosita_superset_v1_stars estars ON estars.gaia_dr3_source_id = g3.source_id
    WHERE (c2g3.version_id = 21) AND /* control version! */
     (c2g3.best is true) AND /* and enforce unique-ish crossmatch */
     estars.xmatch_method = 'HamStar' AND
     estars.xmatch_version = 'v1.1.1' AND
     estars.xmatch_metric > 0.0 AND
     g3.source_phot_g_mean_mag BETWEEN 13.0 and 22.0
    ;

    See cadence logic in post_process() below.

    xmatch_metric now spans the range [0.2:1.0]

    In the query below we use xmatch_metric > 0.0 - i.e. no cut on this param
    """

    # name = 'mwm_erosita_stars'
    name = "mwm_erosita_stars_boss"
    category = "science"
    instrument = "BOSS"
    program = "mwm_erosita"
    mapper = "MWM"
    can_offset = True

    faintest_cadence = "dark_flexible_2x1"
    # faintest_priority = 1910
    flat_priority = 1920

    def build_query(self, version_id, query_region=None):
        bright_bright_limit = 13
        cadence = peewee.Case(
            None,
            (
                (Gaia_DR3.phot_g_mean_mag < 17.0, "bright_1x1"),
                (Gaia_DR3.phot_g_mean_mag < 18.0, "dark_flexible_2x1"),
            ),
            self.faintest_cadence,
        ).cast("text")

        # priority = (
        #     EROSITASupersetv1Stars.xmatch_flags - 1 +
        #     peewee.Case(
        #         None,
        #         (
        #             (Gaia_DR3.phot_g_mean_mag < 17.0, 2400),
        #         ),
        #         self.faintest_priority
        #     ).cast('integer')
        # )
        priority = peewee.Value(self.flat_priority).cast("integer")

        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag.alias("gaia_g"),
                Gaia_DR3.phot_bp_mean_mag.alias("bp"),
                Gaia_DR3.phot_rp_mean_mag.alias("rp"),
                Gaia_DR3.parallax,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                cadence.alias("cadence"),
                priority.alias("priority"),
                EROSITASupersetv1Stars.target_priority,
                EROSITASupersetv1Stars.ero_detuid,
                EROSITASupersetv1Stars.xmatch_metric,
                EROSITASupersetv1Stars.xmatch_flags,
                EROSITASupersetv1Stars.ero_flux,
                EROSITASupersetv1Stars.ero_ra,
                EROSITASupersetv1Stars.ero_dec,
                EROSITASupersetv1Stars.opt_ra,
                EROSITASupersetv1Stars.opt_dec,
            )
            .distinct(CatalogToGaia_DR3.catalogid)
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(
                EROSITASupersetv1Stars,
                on=(Gaia_DR3.source_id == EROSITASupersetv1Stars.gaia_dr3_source_id),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                EROSITASupersetv1Stars.xmatch_method == "HamStar",
                EROSITASupersetv1Stars.xmatch_version == "v1.1.1",
                EROSITASupersetv1Stars.xmatch_metric > 0.0,
                Gaia_DR3.phot_g_mean_mag > bright_bright_limit,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.
        #

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_EROSITA_Compact_Boss_Carton(BaseCarton):
    """MWM eROSITA Compact-Objects (unified for v1)

    Owner: eROSITA Compact objects WG Axel Schwope

    Shorthand name: mwm_erosita_compact_boss

    What is it?: eROSITA selected pointlike X-ray sources with
    likely Gaia counterpart (currently drawn from DR3),
    with high likelihood of being a galactic object
    according to proper motion and parallax,
    and not being an ordinary coronal emitter,
    also being (optically) reasonably faint,
    taking together being a candidate of a compact binary,
    i.e. accreting compact object.
    The original selection criteria that were applied
    to all X-ray/Gaia pairs are described on page Compact Binaries.
    These form the erosita_superset_v1_compactobjects.

    Simplified Description of selection criteria:
    "Select all targets from erosita_superset_compactobjects
    that have minimal distance to eROSITA eRASS:3 position
    and G > 13mag. All selected targets will receive BOSS spectroscopy"

    Wiki page: link or N/A

    Additional source catalogs needed: None

    Additional cross-matching needed: None

    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    minimum exposure is 1x1 for bright 16<g < 17 mag),
    1x2 for medium (17<g<18)
    Fainter targets have 1x2 in the normal version of the cvarton,
    but have 2x2 in the _deep version of this carton.

    Pseudo SQL (optional):

    select * from erosita_superset_v1_compactobjects
    where xmatch_method = 'NWAY_CV' AND
    xmatch_version = '20Oct2022'

    reveals 11113 targets for 11087 X-ray sources

    Implementation:
    SELECT c.catalogid, c.ra, c.dec, gaia.phot_g_mean_mag,
    gaia.phot_bp_mean_mag, gaia.phot_rp_mean_mag, gaia.parallax,
    gaia.pmra, gaia.pmdec,
    estars.target_priority, estars.xmatch_metric,
    estars.ero_flux, estars.ero_ra, estars.ero_dec,
    estars.opt_ra, estars.opt_dec
    FROM catalog c
    INNER JOIN catalog_to_gaia_dr3_source c2g3 USING (catalogid)
    INNER JOIN gaia_dr3_source gaia ON gaia.source_id = c2g3.target_id
    INNER JOIN erosita_superset_v1_compactobjects estars
        ON estars.gaia_dr3_source_id = gaia.source_id
    WHERE (c2g3.version_id = 31) AND /* control version! */
    (c2g3.best is true) AND /* and enforce unique-ish crossmatch */
    estars.xmatch_method = 'NWAY_CV' AND
    estars.xmatch_version = '20Oct2022'
    limit 10;
    """

    name = "mwm_erosita_compact_boss"
    category = "science"
    instrument = "BOSS"
    program = "mwm_erosita"
    mapper = "MWM"
    can_offset = True
    faintest_cadence = "dark_flexible_2x2"
    flat_priority = 1810

    def build_query(self, version_id, query_region=None):
        cadence = peewee.Case(
            None,
            (
                (Gaia_DR3.phot_g_mean_mag < 17.0, "bright_1x1"),
                (Gaia_DR3.phot_g_mean_mag < 18.0, "dark_flexible_2x1"),
            ),
            self.faintest_cadence,
        ).cast("text")

        # priority = peewee.Case(
        #     None,
        #     (
        #         (Gaia_DR3.phot_g_mean_mag < 17.0, 2400),
        #     ),
        #     self.faintest_priority
        # ).cast('integer')
        priority = peewee.Value(self.flat_priority).cast("integer")

        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag.alias("gaia_g"),
                Gaia_DR3.phot_bp_mean_mag.alias("bp"),
                Gaia_DR3.phot_rp_mean_mag.alias("rp"),
                Gaia_DR3.parallax,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                cadence.alias("cadence"),
                priority.alias("priority"),
                EROSITASupersetv1Compactobjects.target_priority,
                EROSITASupersetv1Compactobjects.ero_detuid,
                EROSITASupersetv1Compactobjects.xmatch_metric,
                EROSITASupersetv1Compactobjects.ero_flux,
                EROSITASupersetv1Compactobjects.ero_ra,
                EROSITASupersetv1Compactobjects.ero_dec,
                EROSITASupersetv1Compactobjects.opt_ra,
                EROSITASupersetv1Compactobjects.opt_dec,
            )
            .distinct(CatalogToGaia_DR3.catalogid)
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(
                EROSITASupersetv1Compactobjects,
                on=(Gaia_DR3.source_id == EROSITASupersetv1Compactobjects.gaia_dr3_source_id),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                EROSITASupersetv1Compactobjects.xmatch_method == "NWAY_CV",
                EROSITASupersetv1Compactobjects.xmatch_version == "20Oct2022",
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.
        #
        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_EROSITA_Compact_Boss_Shallow_Carton(MWM_EROSITA_Compact_Boss_Carton):
    name = "mwm_erosita_compact_boss_shallow"
    faintest_cadence = "dark_flexible_2x1"
    faintest_priority = 1812


class MWM_EROSITA_Compact_Boss_D3_Carton(MWM_EROSITA_Compact_Boss_Carton):
    name = "mwm_erosita_compact_boss_d3"
    faintest_cadence = "dark_flexible_3x1"
    faintest_priority = 1811



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-06-05
# @Filename: mwm_cb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    HECATE_1_1,
    MILLIQUAS_7_7,
    UVOT_SSC_1,
    XMM_OM_SUSS_5_0,
    BailerJonesEDR3,
    Catalog,
    CatalogToGaia_DR3,
    CatalogToMILLIQUAS_7_7,
    CatalogToUVOT_SSC_1,
    CatalogToXMM_OM_SUSS_5_0,
    Gaia_DR3,
    Galex_GR7_Gaia_DR3,
    GUVCat,
)

from target_selection import log
from target_selection.cartons import BaseCarton
from target_selection.cartons.tools import create_table_as


class MWM_CB_GALEX_Mag_boss_Carton(BaseCarton):
    """Magnitude-limited GALEX FUV / Gaia selection.

    Simplified Description of selection criteria:
        Uses Nicola Gentile Fusillo's Gaia/GALEX cross-match (which should be already ingested).
        Select objects that lie below the main-sequence in an FUV/optical absolute magnitude-color
        diagram. Remove AGN interlopers. Remove the LMC and SMC, where a lot of hot young stars
        mimic the properties of this target selection. For more details, see here.

    Pseudo-code:
        Use the following short-hands:
            pm = sqrt(pmra*pmra+pmdec*pmdec)
            pm_error = sqrt((pmra_error*pmra/pm)*(pmra_error*pmra/pm)+
                            (pmdec_error*pmdec/pm)*(pmdec_error*pmdec/pm))
            logpmdpm = log10(pm/pm_error)
            poe = parallax / parallax_error
        Then combine the following cuts with &&
            fuv_mag + 5*log10(parallax/1000) + 5 > 1.5 + 1.28*(fuv_mag - phot_g_mean_mag)
            fuv_mag + 5*log10(parallax/1000) + 5 > 2.5*(fuv_mag - phot_g_mean_mag) - 0.5
            fuv_magerr < 0.3
            !(logpmdpm < 0.301 && poe > -1.75*logpmdpm - 2.8 && poe <  1.75*logpmdpm + 2.8 ||
              (logpmdpm-0.301)*(logpmdpm-0.301) / (0.33*0.33) + (poe * poe) / (3.2* 3.2)<=1)
            !((sqrt( pow(l-280.3,2) + 2*pow(b+33.0,2) ) < 8) ||
              sqrt(0.5*square(l-301) + 2*square(b+43.3) ) < 5)

    """

    name = "mwm_cb_galex_mag_boss"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = None  # cadence is set in peewee.Case() below
    program = "mwm_cb"
    priority = 1820
    can_offset = True

    # cadence is set in peewee.Case() below
    # For cadence:
    # G < =16, bright_flexible_2x1
    # G > 16, dark_flexible_2x1
    def build_query(self, version_id, query_region=None):
        G3 = Gaia_DR3
        pm = fn.sqrt(G3.pmra * G3.pmra + G3.pmdec * G3.pmdec)
        pm_error = fn.sqrt(
            (G3.pmra_error * G3.pmra / pm) * (G3.pmra_error * G3.pmra / pm)
            + (G3.pmdec_error * G3.pmdec / pm) * (G3.pmdec_error * G3.pmdec / pm)
        )
        logpmdpm = fn.log(pm / pm_error)
        poe = G3.parallax / G3.parallax_error

        fuv_mag = GUVCat.fuv_mag
        Gm = G3.phot_g_mean_mag
        parallax = G3.parallax
        logpar = fn.log(fn.abs(parallax / 1000))
        ll = G3.l
        bb = G3.b

        fuv_cut = ((fuv_mag + 5 * logpar + 5) > (1.5 + 1.28 * (fuv_mag - Gm))) & (
            (fuv_mag + 5 * logpar + 5) > (2.5 * (fuv_mag - Gm) - 0.5)
        )

        astro_cut1 = ~(
            (
                (logpmdpm < 0.301)
                & (poe > (-1.75 * logpmdpm - 2.8))
                & (poe < (1.75 * logpmdpm + 2.8))
            )
            | (
                (
                    (logpmdpm - 0.301) * (logpmdpm - 0.301) / (0.33 * 0.33)
                    + (poe * poe) / (3.2 * 3.2)
                )
                <= 1
            )
        )

        astro_cut2 = ~(
            (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
            | (fn.sqrt(0.5 * fn.pow(ll - 301, 2) + 2 * fn.pow(bb + 43.3, 2)) < 5)
        )

        cadence = peewee.Case(
            None, ((Gm <= 16, "bright_flexible_2x1"), (Gm > 16, "dark_flexible_2x1"))
        )

        query = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                G3.source_id,
                ll,
                bb,
                parallax,
                Gm,
                fuv_mag,
                cadence.alias("cadence"),
            )
            .join(
                Galex_GR7_Gaia_DR3,
                on=(Gaia_DR3.source_id == Galex_GR7_Gaia_DR3.gaia_edr3_source_id),
            )
            .join(GUVCat, on=(GUVCat.objid == Galex_GR7_Gaia_DR3.galex_objid))
            .join_from(Gaia_DR3, CatalogToGaia_DR3)
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                fuv_cut,
                fuv_mag > -999,
                GUVCat.fuv_magerr < 0.3,
                astro_cut1,
                astro_cut2,
            )
            .order_by(CatalogToGaia_DR3.catalogid, Galex_GR7_Gaia_DR3.galex_separation)
            .distinct(CatalogToGaia_DR3.catalogid)
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_CB_GALEX_Vol_boss_Carton(BaseCarton):
    """Volume-limited GALEX FUV / Gaia selection.

    Simplified Description of selection criteria:
        Uses Nicola Gentile Fusillo's Gaia/GALEX cross-match (which should be already ingested).
        Select objects that lie below the main-sequence in an NUV/optical absolute magnitude-color
        diagram. Remove known AGN & galaxies from published catalogues. Remove AGN interlopers by
        parallax / proper motion. Remove the LMC and SMC, where a lot of hot young stars mimic the
        properties of this target selection. Makes use of milliquas (million quasar catalog
        https://quasars.org/milliquas.htm) and HECATE (Heraklion Extragalactic Catalog -
        https://hecate.ia.forth.gr) to remove contaminants.

    Pseudo-code:
        1. Eliminate AGN - 80 632 559
        Condition: 1" coordinate match, select all objects that do not match an entry in
        MILLIQUAS_7_7 or that match and have QPCT!=100.

        2. Eliminate Galaxies - 77 074 394
        Condition: Eliminate targets that are within 60*R1 of a HECATE Galaxy

        3. Keep matches with GAIA-GALEX coordinate separation below 2.5" - 70 608 597
        Condition: GALEX_Separation<2.5

        4. Valid nuv_mags and nuv error - 30 204 026
        Condition: nuv_mag>-999 && nuv_magerr<0.2 && nuv_magerr>0

        5. HRD/CCD cut - 2 012 017
        Condition: ((nuv_mag-5*log10(r_med_geo)+5>2*(nuv_mag-phot_g_mean_mag)+0.7)
        or(nuv_mag-5*log10(abs(1000/parallax))+5>2*(nuv_mag-phot_g_mean_mag)+0.7)) and
        (nuv_mag-phot_g_mean_mag<=1.6*bp_rp+1.2)"

        6. Astrometry cut - 1 031 545
        Condition:
        (NOT(((log10(sqrt(pmra*pmra+pmdec*pmdec)/sqrt((pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))
        *(pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))+(pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*
        pmdec))*(pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))))<0.301 AND
        (parallax/parallax_error)>-1.75*log10(sqrt(pmra*pmra+pmdec*pmdec)/
        sqrt((pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))*(pmra_error*pmra/
        sqrt(pmra*pmra+pmdec*pmdec))+(pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))*
        (pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))))-2.8 AND
        (parallax/parallax_error)<1.75*log10(sqrt(pmra*pmra+pmdec*pmdec)/
        sqrt((pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))*(pmra_error*pmra/
        sqrt(pmra*pmra+pmdec*pmdec))+(pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))*
        (pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))))+2.8) OR
        (log10(sqrt(pmra*pmra+pmdec*pmdec)/sqrt((pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))*
        (pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))+(pmdec_error*pmdec/
        sqrt(pmra*pmra+pmdec*pmdec))*(pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))))-0.301)*
        (log10(sqrt(pmra*pmra+pmdec*pmdec)/sqrt((pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))*
        (pmra_error*pmra/sqrt(pmra*pmra+pmdec*pmdec))+(pmdec_error*pmdec/
        sqrt(pmra*pmra+pmdec*pmdec))*(pmdec_error*pmdec/sqrt(pmra*pmra+pmdec*pmdec))))-0.301)/
        (0.33*0.33)+(parallax/parallax_error)*(parallax/parallax_error)/(3.2*3.2)<=1)))

        7. Distance cut - 564 537
        Condition: (r_med_geo<=1500|| abs(1000/parallax)<=1500)

        8a. Eliminate GAIA duplicates - 408 446
        Condition: Sort result from step 7 from lowest to largest matching distance and
        keep only closest match for each GAIA source_id

        8b. Eliminate GALEX duplicates - 560 021
        Condition: Sort result from step 7 from lowest to largest matching distance and
        keep only closest match for each GALEX ID

        9. Combine unique catalogues, keeping only single detections both ways - 405 156
        Condition: Match result catalogues from steps 8a and 8b via GAIA source_id -> Numbers in
        agreement with what is shown in uvex_selection_final_numbers.txt

        10. Remove objects in Magellanic Clouds - 402 832
        Condition: !((sqrt( pow(l-280.3,2) + 2*pow(b+33.0,2) ) < 8) || sqrt(0.5*square(l-301) +
                                                                            2*square(b+43.3) ) < 5)

    """

    name = "mwm_cb_galex_vol_boss"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = None  # cadence is set in peewee.Case() below
    program = "mwm_cb"
    priority = 1820
    can_offset = True

    # cadence is set in peewee.Case() below
    # For cadence:
    # G <= 16, bright_flexible_2x1
    # G > 16, dark_flexible_2x1
    def build_query(self, version_id, query_region=None):
        self.database.execute_sql('SET work_mem = "10GB";')

        GR7 = Galex_GR7_Gaia_DR3
        G3 = Gaia_DR3

        nuv_cut = (
            GR7.select(GR7.gaia_edr3_source_id, GR7.galex_objid, GR7.galex_separation, GR7.nuv_mag)
            .where(
                GR7.nuv_mag > -999,
                GR7.nuv_magerr < 0.2,
                GR7.nuv_magerr > 0,
                GR7.galex_separation <= 2.5,
            )
            .cte("nuv_cut")
        )

        Gm = G3.phot_g_mean_mag
        cadence = peewee.Case(
            None, ((Gm <= 16, "bright_flexible_2x1"), (Gm > 16, "dark_flexible_2x1"))
        )

        gaia_query = (
            G3.select(
                G3.source_id,
                G3.pmra,
                G3.pmdec,
                G3.ra,
                G3.dec,
                G3.l,
                G3.b,
                G3.parallax,
                G3.parallax_error,
                G3.pmra_error,
                G3.pmdec_error,
                G3.phot_g_mean_mag,
                G3.bp_rp,
                cadence.alias("cadence"),
                nuv_cut.c.nuv_mag,
                nuv_cut.c.galex_objid,
                nuv_cut.c.galex_separation,
                BailerJonesEDR3.r_med_geo,
            )
            .join(nuv_cut, on=(nuv_cut.c.gaia_edr3_source_id == G3.source_id))
            .join(BailerJonesEDR3, on=(BailerJonesEDR3.source_id == G3.source_id))
            .with_cte(nuv_cut)
        )

        log.debug("Collecting Gaia-GALEX information ...")

        cb_temp1, _ = create_table_as(
            gaia_query, "cb_temp1", temporary=True, database=self.database, overwrite=True
        )

        log.debug(f"Gaia-Galex NUV cut returned {cb_temp1.select().count():,} rows.")

        sqrt = fn.sqrt
        pow = fn.pow
        log10 = fn.log10
        pabs = fn.abs

        pmra = cb_temp1.c.pmra
        pmdec = cb_temp1.c.pmdec
        pmra_error = cb_temp1.c.pmra_error
        pmdec_error = cb_temp1.c.pmdec_error
        parallax = cb_temp1.c.parallax
        parallax_error = cb_temp1.c.parallax_error

        pm = sqrt(pmra * pmra + pmdec * pmdec)
        pm_error = sqrt(
            (pmra_error * pmra / pm) * (pmra_error * pmra / pm)
            + (pmdec_error * pmdec / pm) * (pmdec_error * pmdec / pm)
        )
        perror2 = (parallax / parallax_error) * (parallax / parallax_error)

        # Astro cut
        astro_cut = ~(
            (
                (log10(pm / pm_error) < 0.301)
                & ((parallax / parallax_error) > -1.75 * log10(pm / pm_error) - 2.8)
                & ((parallax / parallax_error) < 1.75 * log10(pm / pm_error) + 2.8)
            )
            | (
                (log10(pm / pm_error) - 0.301) * (log10(pm / pm_error) - 0.301) / (0.33**2)
                + perror2 / (3.2**2)
                <= 1
            )
        )

        nuv_mag = cb_temp1.c.nuv_mag
        r_med_geo = cb_temp1.c.r_med_geo
        phot_g_mean_mag = cb_temp1.c.phot_g_mean_mag
        bp_rp = cb_temp1.c.bp_rp

        d_pc = pabs(1000 / parallax)
        hrd_bj = (nuv_mag - 5 * log10(r_med_geo) + 5) > (2 * (nuv_mag - phot_g_mean_mag) + 0.7)
        hrd_gaia = (nuv_mag - 5 * log10(d_pc) + 5) > (2 * (nuv_mag - phot_g_mean_mag) + 0.7)

        # HRD/CCD cut
        hrd_cut = (hrd_bj | hrd_gaia) & (nuv_mag - phot_g_mean_mag <= (1.6 * bp_rp + 1.2))

        # Distance cut
        distance_cut = (d_pc <= 1500) | (cb_temp1.c.r_med_geo <= 1500)

        # MC cut
        ll = cb_temp1.c.l
        bb = cb_temp1.c.b
        mc_cut = ~(
            (sqrt(pow(ll - 280.3, 2) + 2 * pow(bb + 33.0, 2)) < 8)
            | (sqrt(0.5 * pow(ll - 301, 2) + 2 * pow(bb + 43.3, 2)) < 5)
        )

        cb_temp2_query = cb_temp1.select(
            cb_temp1.c.source_id,
            cb_temp1.c.ra,
            cb_temp1.c.dec,
            cb_temp1.c.galex_objid,
            cb_temp1.c.galex_separation,
            cb_temp1.c.cadence,
        ).where(astro_cut, hrd_cut, distance_cut, mc_cut)

        log.debug("Applying astrometric and colour cuts ...")

        cb_temp2, _ = create_table_as(
            cb_temp2_query, "cb_temp2", temporary=True, database=self.database, overwrite=True
        )

        log.debug(f"Astrometric and colour cuts returned {cb_temp2.select().count():,} rows.")

        # Remove AGNs, although this doesn't seem to actually reject any.
        agn_query = (
            cb_temp2.select(
                CatalogToGaia_DR3.catalogid,
                cb_temp2.c.source_id,
                cb_temp2.c.ra,
                cb_temp2.c.dec,
                cb_temp2.c.galex_objid,
                cb_temp2.c.galex_separation,
                cb_temp2.c.cadence,
            )
            .join(CatalogToGaia_DR3, on=(CatalogToGaia_DR3.target_id == cb_temp2.c.source_id))
            .join(
                CatalogToMILLIQUAS_7_7,
                on=(CatalogToGaia_DR3.catalogid == CatalogToMILLIQUAS_7_7.catalogid),
                join_type=peewee.JOIN.LEFT_OUTER,
            )
            .join(
                MILLIQUAS_7_7,
                on=(MILLIQUAS_7_7.pk == CatalogToMILLIQUAS_7_7.target_id),
                join_type=peewee.JOIN.LEFT_OUTER,
            )
            .where(
                CatalogToMILLIQUAS_7_7.best.is_null()
                | ((CatalogToMILLIQUAS_7_7.best >> True) & (MILLIQUAS_7_7.qpct != 100))
            )
            .where(CatalogToGaia_DR3.best >> True, CatalogToGaia_DR3.version_id == version_id)
        )

        cb_temp3, _ = create_table_as(
            agn_query,
            "cb_temp3",
            temporary=True,
            database=self.database,
            overwrite=True,
            indices=["q3c_ang2ipix(ra,dec)"],
        )

        log.debug(f"Remaining after rejecting AGNs: {cb_temp3.select().count():,} rows")

        # Select targets without nearby galaxies. First create a list of galaxies
        # within hecate_1_1.r1 arcmin of each of the targets.
        gal_query = (
            cb_temp3.select(cb_temp3.c.source_id)
            .join(HECATE_1_1, join_type=peewee.JOIN.CROSS)
            .where(
                fn.q3c_join(
                    HECATE_1_1.ra,
                    HECATE_1_1.dec,
                    cb_temp3.c.ra,
                    cb_temp3.c.dec,
                    HECATE_1_1.r1 / 60.0,
                ),
                HECATE_1_1.r1 != "NaN",
            )
            .cte("hecate_query", materialized=True)
        )

        # Now keep only targets that are not in the previous list.
        gal_free_query = (
            cb_temp3.select(
                cb_temp3.c.catalogid,
                cb_temp3.c.source_id,
                cb_temp3.c.ra,
                cb_temp3.c.dec,
                cb_temp3.c.galex_objid,
                cb_temp3.c.galex_separation,
                cb_temp3.c.cadence,
            )
            .join(
                gal_query,
                on=(gal_query.c.source_id == cb_temp3.c.source_id),
                join_type=peewee.JOIN.LEFT_OUTER,
            )
            .where(gal_query.c.source_id.is_null())
            .with_cte(gal_query)
        )

        cb_temp4, _ = create_table_as(
            gal_free_query, "cb_temp4", temporary=True, database=self.database, overwrite=True
        )

        log.debug(
            "Remaining after filtering for nearby galaxies: "
            f"{cb_temp4.select().count():,} rows."
        )

        # Get unique Gaia and GALEX targets with minimum separaton, independently, and keep
        # source_ids that are in both groups.

        # First we create a subquery partitioned by source_id and rank each row by separation.
        gaia_ranked = cb_temp4.select(
            cb_temp4.c.source_id,
            fn.rank()
            .over(
                partition_by=[cb_temp4.c.source_id], order_by=[cb_temp4.c.galex_separation.asc()]
            )
            .alias("sep_rank"),
        ).alias("gaia_ranked")

        # Now we select only the entry with sep_rank=1 (minimum galex_separation).
        gaia_unique = (
            cb_temp4.select(cb_temp4.c.source_id)
            .join(gaia_ranked, on=(cb_temp4.c.source_id == gaia_ranked.c.source_id))
            .where(gaia_ranked.c.sep_rank == 1)
            .distinct(cb_temp4.c.source_id)
            .alias("gaia_unique")
        )

        # Do the same for galex_objid but keep the Gaia source since that's the one we'll join on.
        galex_ranked = cb_temp4.select(
            cb_temp4.c.source_id,
            fn.rank()
            .over(
                partition_by=[cb_temp4.c.galex_objid], order_by=[cb_temp4.c.galex_separation.asc()]
            )
            .alias("sep_rank"),
        ).alias("galex_ranked")

        galex_unique = (
            cb_temp4.select(cb_temp4.c.source_id)
            .join(galex_ranked, on=(cb_temp4.c.source_id == galex_ranked.c.source_id))
            .where(galex_ranked.c.sep_rank == 1)
            .distinct(cb_temp4.c.galex_objid)
            .alias("galex_unique")
        )

        # Now keep source_ids that exist in both subsamples. Note we need to distinct on
        # source_id again because galex_unique has duplicate source_ids (different galex_objid
        # can have the same source_id).
        unique = (
            cb_temp4.select(
                cb_temp4.c.catalogid,
                cb_temp4.c.source_id,
                cb_temp4.c.ra,
                cb_temp4.c.dec,
                cb_temp4.c.galex_objid,
                cb_temp4.c.galex_separation,
                cb_temp4.c.cadence,
            )
            .join(gaia_unique, on=(gaia_unique.c.source_id == cb_temp4.c.source_id))
            .join(galex_unique, on=(galex_unique.c.source_id == cb_temp4.c.source_id))
            .distinct(cb_temp4.c.source_id)
        )

        log.debug("Removing duplicates ...")

        return unique


class MWM_CB_XMMOM_boss_Carton(BaseCarton):
    """Volume-limited XMM-OM / Gaia selection.

    Simplified Description of selection criteria:
        Will use matching XMM-OM / Gaia DR3 sources and slightly revised color cuts compared to
        V0.5. Select objects that lie below the main-sequence in an NUV/optical absolute
        magnitude-color diagram. Remove known AGN & galaxies from published catalogues. Remove AGN
        interlopers by parallax / proper motion. Remove the LMC and SMC, where a lot of hot young
        stars mimic the properties of this target selection. Makes use of milliquas (million quasar
        catalog https://quasars.org/milliquas.htm) and HECATE (Heraklion Extragalactic Catalog -
        https://hecate.ia.forth.gr).

    Pseudo-code:
        * Preparation:
            Trim milliquas catalog
                java -jar stilts.jar tpipe in=milliquas.fits cmd='select QPCT==100'
                    out=milliquas_good1.fits
            Reduce file size of XMMOM input by selecting only used columns and by applying
            quality cuts:
                java -jar stilts.jar tpipe in=XMM-OM-SUSS5.0.fits out=xmm_reduced1.fits
                    cmd="select UVM2_AB_MAG>=-100" cmd="keepcols 'SRCNUM OBSID RA DEC UVW1_AB_MAG
                    UVW2_AB_MAG UVM2_AB_MAG UVW1_QUALITY_FLAG_ST UVW2_QUALITY_FLAG_ST
                    UVM2_QUALITY_FLAG_ST UVW1_SIGNIF UVW2_SIGNIF UVM2_SIGNIF'"
        * Gaia match within 1.5" and Bailer-Jones distances
            java -jar stilts.jar cdsskymatch cdstable=I/350/gaiaedr3 find=all
                icmd='select (!(equals(substring(UVM2_QUALITY_FLAG_ST,0,1),\"T\")||
                        (equals(substring(UVM2_QUALITY_FLAG_ST,1,2),\"T\")&&UVM2_SIGNIF<10)||
                        (equals(substring(UVM2_QUALITY_FLAG_ST,2,3),\"T\")&&UVM2_SIGNIF<10)||
                        equals(substring(UVM2_QUALITY_FLAG_ST,6,7),\"T\")||
                        equals(substring(UVM2_QUALITY_FLAG_ST,7,8),\"T\")||
                        equals(substring(UVM2_QUALITY_FLAG_ST,8,9),\"T\")))'
                    in=xmm_reduced1.fits ra= dec=DEC radius=1.5 blocksize=50000
                    out=xmm_gaia.fits
            java -jar stilts.jar cdsskymatch cdstable=I/352/gedr3dis find=best
                in=xmm_gaia.fits ra=ra_cds dec=dec_cds radius=0.001
                blocksize=50000 out=xmm_gaia_dist.fits
        * Eliminate known AGN, i.e. all milliquas entries with 100% probability being an AGN
          and eliminate UV sources in nearby galaxies
            java -jar -Xmx16G stilts.jar tmatch2 in1=xmm_gaia_dist.fits in2=milliquas_good1.fits
                matcher=sky params=1.0 values1='ra_cds dec_cds' values2='RA DEC'
                join=1not2 find=all out=xmm_gaia_dist_noAGN.fits
            java -jar -Xmx16G stilts.jar tmatch2 in1=xmm_gaia_dist_noAGN.fits in2=HECATE_v1.1.fits
                matcher=skyerr params=10 values1='ra_cds dec_cds 0.0001' values2='RA DEC 60*R1'
                join=1not2 find=all out=reduced_XMM_all_data.fits
        * Select entries with maximum separation between Gaia and XMM-OM position of 1.5 arcsec
        * Select objects with an UV excess via two cuts, one in the in the NUV-HRD,
          the other in an UV-to-optical color-color diagram
          - UVM2_AB_MAG>-999 && UVM2_AB_MAG-5*log10(rgeo)+5>2*(UVM2_AB_MAG-phot_g_mean_mag)+0.7 ||
            (UVM2_AB_MAG-5*log10(parallax/1000)+5>2*(UVM2_AB_MAG-phot_g_mean_mag)+0.7)
          - UVM2_AB_mag - phot_g_mean_mag < 1.6 * (bp-rp) + 1.2
        * Astrometric cut: to remove AGN contaminants:
            !(logpmdpm < 0.301 && poe > -1.75*logpmdpm - 2.8 && poe <  1.75*logpmdpm + 2.8 ||
            (logpmdpm-0.301)*(logpmdpm-0.301) / (0.33*0.33) + (poe * poe) / (3.2* 3.2)<=1)
        * Distance cut: r_med_geo<=1500 ||  1000/parallax <= 1500
        * Eliminate duplicates both ways (Gaia and XMMOM), keep always nearest
          Gaia - XMMOM match only.

    """

    name = "mwm_cb_xmmom_boss"
    mapper = "MWM"
    category = "science"
    instrument = "BOSS"
    cadence = None  # cadence is set in peewee.Case() below
    program = "mwm_cb"
    priority = 1820
    can_offset = True

    # cadence is set in peewee.Case() below
    # For cadence:
    # G <= 16, bright_flexible_2x1
    # G > 16, dark_flexible_2x1
    def build_query(self, version_id, query_region=None):
        G3 = Gaia_DR3
        pm = fn.sqrt(G3.pmra * G3.pmra + G3.pmdec * G3.pmdec)
        pm_error = fn.sqrt(
            (G3.pmra_error * G3.pmra / pm) * (G3.pmra_error * G3.pmra / pm)
            + (G3.pmdec_error * G3.pmdec / pm) * (G3.pmdec_error * G3.pmdec / pm)
        )
        logpmdpm = fn.log(pm / pm_error)
        poe = G3.parallax / G3.parallax_error

        Gm = G3.phot_g_mean_mag
        bp_rp = G3.bp_rp
        parallax = G3.parallax
        logpar = fn.log(fn.abs(1000 / parallax))

        r_med_geo = BailerJonesEDR3.r_med_geo

        XMM = XMM_OM_SUSS_5_0
        uvm2 = XMM.uvm2_ab_mag

        qcut = ~(
            (fn.substring(XMM.uvm2_quality_flag_st, 1, 1) == "T")
            | ((fn.substring(XMM.uvm2_quality_flag_st, 2, 1) == "T") & (XMM.uvm2_signif < 10))
            | ((fn.substring(XMM.uvm2_quality_flag_st, 3, 1) == "T") & (XMM.uvm2_signif < 10))
            | (fn.substring(XMM.uvm2_quality_flag_st, 7, 1) == "T")
            | (fn.substring(XMM.uvm2_quality_flag_st, 8, 1) == "T")
            | (fn.substring(XMM.uvm2_quality_flag_st, 9, 1) == "T")
        )

        color_cuts = [
            (uvm2 >= -100),
            (
                ((uvm2 - 5 * fn.log10(r_med_geo) + 5) > (2 * (uvm2 - Gm) + 0.7))
                | ((uvm2 - 5 * logpar + 5) > (2 * (uvm2 - Gm) + 0.7))
            ),
            (uvm2 - Gm < 1.6 * bp_rp + 1.2),
        ]

        astro_cut = ~(
            (logpmdpm < 0.301) & (poe > (-1.75 * logpmdpm - 2.8)) & (poe < (1.75 * logpmdpm + 2.8))
            | (
                (
                    (logpmdpm - 0.301) * (logpmdpm - 0.301) / (0.33 * 0.33)
                    + (poe * poe) / (3.2 * 3.2)
                )
                <= 1
            )
        )

        cadence = peewee.Case(
            None, ((Gm <= 16, "bright_flexible_2x1"), (Gm > 16, "dark_flexible_2x1"))
        )

        # MC cut
        ll = G3.l
        bb = G3.b
        mc_cut = ~(
            (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
            | (fn.sqrt(0.5 * fn.pow(ll - 301, 2) + 2 * fn.pow(bb + 43.3, 2)) < 5)
        )

        q1 = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                G3.source_id,
                G3.ra,
                G3.dec,
                CatalogToXMM_OM_SUSS_5_0.distance.alias("separation"),
                XMM_OM_SUSS_5_0.srcnum.alias("srcnum"),
                cadence.alias("cadence"),
            )
            .join_from(
                Gaia_DR3, BailerJonesEDR3, on=(Gaia_DR3.source_id == BailerJonesEDR3.source_id)
            )
            .join_from(Gaia_DR3, CatalogToGaia_DR3)
            .join_from(
                CatalogToGaia_DR3,
                CatalogToMILLIQUAS_7_7,
                join_type=peewee.JOIN.LEFT_OUTER,
                on=(CatalogToMILLIQUAS_7_7.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(MILLIQUAS_7_7, join_type=peewee.JOIN.LEFT_OUTER)
            .join_from(
                CatalogToGaia_DR3,
                CatalogToXMM_OM_SUSS_5_0,
                on=(CatalogToGaia_DR3.catalogid == CatalogToXMM_OM_SUSS_5_0.catalogid),
            )
            .join(XMM_OM_SUSS_5_0)
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToXMM_OM_SUSS_5_0.distance <= 1.5 / 3600.0,
                (
                    CatalogToMILLIQUAS_7_7.catalogid.is_null()
                    | ((CatalogToMILLIQUAS_7_7.best >> True) & (MILLIQUAS_7_7.qpct != 100))
                ),
                (r_med_geo <= 1500) | (fn.abs(1000 / parallax) <= 1500),
                qcut,
                *color_cuts,
                astro_cut,
                mc_cut,
            )
        )

        xmm_temp, _ = create_table_as(
            q1,
            "mwm_cb_xmmom_temp",
            temporary=True,
            database=self.database,
            indices=["q3c_ang2ipix(ra,dec)"],
        )

        # Select targets without nearby galaxies. First create a list of galaxies
        # within hecate_1_1.r1 arcmin of each of the targets.
        gal_query = (
            xmm_temp.select(xmm_temp.c.source_id)
            .join(HECATE_1_1, join_type=peewee.JOIN.CROSS)
            .where(
                fn.q3c_join(
                    HECATE_1_1.ra,
                    HECATE_1_1.dec,
                    xmm_temp.c.ra,
                    xmm_temp.c.dec,
                    HECATE_1_1.r1 / 60.0,
                ),
                HECATE_1_1.r1 != "NaN",
            )
            .cte("hecate_query", materialized=True)
        )

        # Now keep only targets that are not in the previous list.
        gal_free_query = (
            xmm_temp.select(
                xmm_temp.c.catalogid,
                xmm_temp.c.source_id,
                xmm_temp.c.srcnum,
                xmm_temp.c.separation,
                xmm_temp.c.ra,
                xmm_temp.c.dec,
                xmm_temp.c.cadence,
            )
            .join(
                gal_query,
                on=(gal_query.c.source_id == xmm_temp.c.source_id),
                join_type=peewee.JOIN.LEFT_OUTER,
            )
            .where(gal_query.c.source_id.is_null())
            .with_cte(gal_query)
        )

        xmm_temp2, _ = create_table_as(
            gal_free_query, "mwm_cb_xmmom_temp2", temporary=True, database=self.database
        )

        # Get unique Gaia and XMMOM targets with minimum separaton, independently, and keep
        # source_ids that are in both groups.

        # First we create a subquery partitioned by source_id and rank each row by separation.
        gaia_ranked = xmm_temp2.select(
            xmm_temp2.c.source_id,
            fn.rank()
            .over(partition_by=[xmm_temp2.c.source_id], order_by=[xmm_temp2.c.separation.asc()])
            .alias("sep_rank"),
        ).alias("gaia_ranked")

        # Now we select only the entry with sep_rank=1 (minimum separation).
        gaia_unique = (
            xmm_temp2.select(xmm_temp2.c.source_id)
            .join(gaia_ranked, on=(xmm_temp2.c.source_id == gaia_ranked.c.source_id))
            .where(gaia_ranked.c.sep_rank == 1)
            .distinct(xmm_temp2.c.source_id)
            .alias("gaia_unique")
        )

        # Do the same for XMMOM id but keep the Gaia source since that's the one we'll join on.
        xmm_ranked = xmm_temp2.select(
            xmm_temp2.c.source_id,
            fn.rank()
            .over(partition_by=[xmm_temp2.c.srcnum], order_by=[xmm_temp2.c.separation.asc()])
            .alias("sep_rank"),
        ).alias("xmm_ranked")

        xmm_unique = (
            xmm_temp2.select(xmm_temp2.c.source_id)
            .join(xmm_ranked, on=(xmm_temp2.c.source_id == xmm_ranked.c.source_id))
            .where(xmm_ranked.c.sep_rank == 1)
            .distinct(xmm_temp2.c.srcnum)
            .alias("xmm_unique")
        )

        # Now keep source_ids that exist in both subsamples. Note we need to distinct on
        # source_id again because xmm_unique has duplicate source_ids (different srcnum
        # can have the same source_id).
        unique = (
            xmm_temp2.select(
                xmm_temp2.c.catalogid,
                xmm_temp2.c.source_id,
                xmm_temp2.c.ra,
                xmm_temp2.c.dec,
                xmm_temp2.c.srcnum,
                xmm_temp2.c.separation,
                xmm_temp2.c.cadence,
            )
            .join(gaia_unique, on=(gaia_unique.c.source_id == xmm_temp2.c.source_id))
            .join(xmm_unique, on=(xmm_unique.c.source_id == xmm_temp2.c.source_id))
            .distinct(xmm_temp2.c.source_id)
        )

        return unique


class MWM_CB_SWIFTUVOT_boss_Carton(BaseCarton):
    """MWM Compact Binaries UV excess 4.

    Simplified Description of selection criteria:
        Will use matching Swift-UVOT / Gaia DR3 sources and slightly revised color cuts compared to
        V0.5. Select objects that lie below the main-sequence in an NUV/optical absolute
        magnitude-color diagram. Remove known AGN & galaxies from published catalogues. Remove AGN
        interlopers by parallax / proper motion. Remove the LMC and SMC, where a lot of hot young
        stars mimic the properties of this target selection. Makes use of milliquas (million quasar
        catalog https://quasars.org/milliquas.htm) and HECATE (Heraklion Extragalactic Catalog -
        https://hecate.ia.forth.gr)  ← ingestion 'done' in the catalog ingestion list on the
        timeline + tasks for v1 page.  Full details, see here.

    Pseudo-code:
        - Target selection analog to mwm_cb_xmmom. Names of input columns in swiftuvot catalog
          slightly different:
        - Quality cuts in input catalog:
            select (!(equals(substring(UVM2_QUALITY_FLAG_BIN,8,9),"1") ||
            (equals(substring(UVM2_QUALITY_FLAG_BIN,7,8),"1") && UVM2_SIGNIF<10) ||
            (equals(substring(UVM2_QUALITY_FLAG_BIN,6,7),"1") && UVM2_SIGNIF<10) ||
             equals(substring(UVM2_QUALITY_FLAG_BIN,3,4),"1") && UVM2_SIGNIF<10) ||
             equals(substring(UVM2_QUALITY_FLAG_BIN,2,3),"1"))"'
        - Gaia match within 1.5" and Bailer-Jones distances (r_med_geo)
        - Eliminate known AGN, i.e. all milliquas entries with 100% probability
          being an AGN and eliminate UV sources in nearby galaxies
            java -jar -Xmx16G stilts.jar tmatch2 in1=xmm_gaia_dist.fits in2=milliquas_good1.fits
              matcher=sky params=1.0 values1='ra_cds dec_cds' values2='RA DEC' join=1not2
              find=all out=swiftuvot_gaia_dist_noAGN.fits
            java -jar -Xmx16G stilts.jar tmatch2 in1=swiftuvot_gaia_dist_noAGN.fits
              in2=HECATE_v1.1.fits matcher=skyerr params=10 values1='ra_cds dec_cds 0.0001'
              values2='RA DEC 60*R1' join=1not2 find=all out=reduced_swift_all_data.fits
        - Select objects with an UV excess via two cuts, one in the in the NUV-HRD, the other
          in an UV-to-optical color-color diagram:
            UVM2_ABMAG>-999 && UVM2_ABMAG-5*log10(r_med_geo)+5>2*(UVM2_ABMAG-phot_g_mean_mag)+0.7
                || (UVM2_ABMAG-5*log10(parallax/1000)+5>2*(UVM2_ABMAG-phot_g_mean_mag)+0.7)
            UVM2_ABMAG - phot_g_mean_mag < 1.6 * (bp-rp) + 1.2
        - Astrometric cut: to remove AGN contaminants:
            !(logpmdpm < 0.301 && poe > -1.75*logpmdpm - 2.8 && poe <  1.75*logpmdpm + 2.8 ||
            (logpmdpm-0.301)*(logpmdpm-0.301) / (0.33*0.33) + (poe * poe) / (3.2* 3.2)<=1)
        - Distance cut: r_med_geo<=1500 ||  1000/parallax <= 1500
        - Eliminate duplicates both ways (Gaia and Swift), keep always nearest
          Gaia - Swiftuvot match only.

    """

    name = "mwm_cb_swiftuvot_boss"
    mapper = "MWM"
    category = "science"
    program = "mwm_cb"
    instrument = "BOSS"
    cadence = None  # cadence is set in peewee.Case() below
    priority = 1820

    # cadence is set in peewee.Case() below
    # For cadence:
    # G <= 16, bright_flexible_2x1
    # G > 16, dark_flexible_2x1
    def build_query(self, version_id, query_region=None):
        G3 = Gaia_DR3
        pm = fn.sqrt(G3.pmra * G3.pmra + G3.pmdec * G3.pmdec)
        pm_error = fn.sqrt(
            (G3.pmra_error * G3.pmra / pm) * (G3.pmra_error * G3.pmra / pm)
            + (G3.pmdec_error * G3.pmdec / pm) * (G3.pmdec_error * G3.pmdec / pm)
        )
        logpmdpm = fn.log(pm / pm_error)
        poe = G3.parallax / G3.parallax_error

        Gm = G3.phot_g_mean_mag
        bp_rp = G3.bp_rp

        parallax = G3.parallax
        logpar = fn.log(fn.abs(1000 / parallax))

        r_med_geo = BailerJonesEDR3.r_med_geo

        uvw2_ab_mag = UVOT_SSC_1.uvw2_ab
        uvm2 = UVOT_SSC_1.uvm2_ab
        uvw1_ab_mag = UVOT_SSC_1.uvw1_ab

        uvm2_quality_flag = UVOT_SSC_1.fuvm2
        uvm2_signif = UVOT_SSC_1.suvm2

        qcut = (
            (uvw2_ab_mag > -100) | (uvw1_ab_mag > -100) | (uvm2 > -100),
            ~(
                (uvm2_quality_flag.bin_and(2**0) > 0)
                | (uvm2_quality_flag.bin_and(2**1) > 0) & (uvm2_signif < 10)
                | (uvm2_quality_flag.bin_and(2**2) > 0) & (uvm2_signif < 10)
                | (uvm2_quality_flag.bin_and(2**5) > 0) & (uvm2_signif < 10)
                | (uvm2_quality_flag.bin_and(2**6) > 0)
            ),
        )

        color_cuts = [
            uvm2 > -100,
            ((uvm2 - 5 * fn.log10(r_med_geo) + 5) > (2 * (uvm2 - Gm) + 0.7))
            | ((uvm2 - 5 * logpar + 5) > (2 * (uvm2 - Gm) + 0.7)),
            (uvm2 - Gm < 1.6 * bp_rp + 1.2),
        ]

        astro_cut = ~(
            (logpmdpm < 0.301) & (poe > (-1.75 * logpmdpm - 2.8)) & (poe < (1.75 * logpmdpm + 2.8))
            | (
                (
                    (logpmdpm - 0.301) * (logpmdpm - 0.301) / (0.33 * 0.33)
                    + (poe * poe) / (3.2 * 3.2)
                )
                <= 1
            )
        )

        cadence = peewee.Case(
            None, ((Gm <= 16, "bright_flexible_2x1"), (Gm > 16, "dark_flexible_2x1"))
        )

        # MC cut
        ll = G3.l
        bb = G3.b
        mc_cut = ~(
            (fn.sqrt(fn.pow(ll - 280.3, 2) + 2 * fn.pow(bb + 33.0, 2)) < 8)
            | (fn.sqrt(0.5 * fn.pow(ll - 301, 2) + 2 * fn.pow(bb + 43.3, 2)) < 5)
        )

        q1 = (
            Gaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                G3.source_id,
                G3.ra,
                G3.dec,
                UVOT_SSC_1.id.alias("uvot_id"),
                CatalogToUVOT_SSC_1.distance.alias("separation"),
                cadence.alias("cadence"),
            )
            .join_from(
                Gaia_DR3, BailerJonesEDR3, on=(Gaia_DR3.source_id == BailerJonesEDR3.source_id)
            )
            .join_from(Gaia_DR3, CatalogToGaia_DR3)
            .join_from(
                CatalogToGaia_DR3,
                CatalogToMILLIQUAS_7_7,
                join_type=peewee.JOIN.LEFT_OUTER,
                on=(CatalogToMILLIQUAS_7_7.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(MILLIQUAS_7_7, join_type=peewee.JOIN.LEFT_OUTER)
            .join_from(
                CatalogToGaia_DR3,
                CatalogToUVOT_SSC_1,
                on=(CatalogToGaia_DR3.catalogid == CatalogToUVOT_SSC_1.catalogid),
            )
            .join(UVOT_SSC_1)
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                (
                    CatalogToMILLIQUAS_7_7.catalogid.is_null()
                    | ((CatalogToMILLIQUAS_7_7.best >> True) & (MILLIQUAS_7_7.qpct != 100))
                ),
                CatalogToUVOT_SSC_1.distance <= 1.5 / 3600.0,
                (r_med_geo <= 1500) | (fn.abs(1000 / parallax) <= 1500),
                *qcut,
                *color_cuts,
                astro_cut,
                mc_cut,
            )
            .order_by(CatalogToGaia_DR3.catalogid, CatalogToUVOT_SSC_1.distance)
            .distinct(CatalogToGaia_DR3.catalogid)
        )

        uvot_temp, _ = create_table_as(
            q1,
            "mwm_cb_uvot_temp",
            temporary=True,
            database=self.database,
            indices=["q3c_ang2ipix(ra,dec)"],
        )

        # Select targets without nearby galaxies. First create a list of galaxies
        # within hecate_1_1.r1 arcmin of each of the targets.
        gal_query = (
            uvot_temp.select(uvot_temp.c.source_id)
            .join(HECATE_1_1, join_type=peewee.JOIN.CROSS)
            .where(
                fn.q3c_join(
                    HECATE_1_1.ra,
                    HECATE_1_1.dec,
                    uvot_temp.c.ra,
                    uvot_temp.c.dec,
                    HECATE_1_1.r1 / 60.0,
                ),
                HECATE_1_1.r1 != "NaN",
            )
            .cte("hecate_query", materialized=True)
        )

        # Now keep only targets that are not in the previous list.
        gal_free_query = (
            uvot_temp.select(
                uvot_temp.c.catalogid,
                uvot_temp.c.source_id,
                uvot_temp.c.uvot_id,
                uvot_temp.c.separation,
                uvot_temp.c.ra,
                uvot_temp.c.dec,
                uvot_temp.c.cadence,
            )
            .join(
                gal_query,
                on=(gal_query.c.source_id == uvot_temp.c.source_id),
                join_type=peewee.JOIN.LEFT_OUTER,
            )
            .where(gal_query.c.source_id.is_null())
            .with_cte(gal_query)
        )

        uvot_temp2, _ = create_table_as(
            gal_free_query, "mwm_cb_uvotom_temp2", temporary=True, database=self.database
        )

        # Get unique Gaia and UVOT targets with minimum separaton, independently, and keep
        # source_ids that are in both groups.

        # First we create a subquery partitioned by source_id and rank each row by separation.
        gaia_ranked = uvot_temp2.select(
            uvot_temp2.c.source_id,
            fn.rank()
            .over(partition_by=[uvot_temp2.c.source_id], order_by=[uvot_temp2.c.separation.asc()])
            .alias("sep_rank"),
        ).alias("gaia_ranked")

        # Now we select only the entry with sep_rank=1 (minimum separation).
        gaia_unique = (
            uvot_temp2.select(uvot_temp2.c.source_id)
            .join(gaia_ranked, on=(uvot_temp2.c.source_id == gaia_ranked.c.source_id))
            .where(gaia_ranked.c.sep_rank == 1)
            .distinct(uvot_temp2.c.source_id)
            .alias("gaia_unique")
        )

        # Do the same for UVOT id but keep the Gaia source since that's the one we'll join on.
        uvot_ranked = uvot_temp2.select(
            uvot_temp2.c.source_id,
            fn.rank()
            .over(partition_by=[uvot_temp2.c.uvot_id], order_by=[uvot_temp2.c.separation.asc()])
            .alias("sep_rank"),
        ).alias("uvot_ranked")

        uvot_unique = (
            uvot_temp2.select(uvot_temp2.c.source_id)
            .join(uvot_ranked, on=(uvot_temp2.c.source_id == uvot_ranked.c.source_id))
            .where(uvot_ranked.c.sep_rank == 1)
            .distinct(uvot_temp2.c.uvot_id)
            .alias("uvot_unique")
        )

        # Now keep source_ids that exist in both subsamples. Note we need to distinct on
        # source_id again because uvot_unique has duplicate source_ids (different uvot_objid
        # can have the same source_id).
        unique = (
            uvot_temp2.select(
                uvot_temp2.c.catalogid,
                uvot_temp2.c.source_id,
                uvot_temp2.c.ra,
                uvot_temp2.c.dec,
                uvot_temp2.c.uvot_id,
                uvot_temp2.c.separation,
                uvot_temp2.c.cadence,
            )
            .join(gaia_unique, on=(gaia_unique.c.source_id == uvot_temp2.c.source_id))
            .join(uvot_unique, on=(uvot_unique.c.source_id == uvot_temp2.c.source_id))
            .distinct(uvot_temp2.c.source_id)
        )

        return unique



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-06-12
# @Filename: bhm_gua.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# isort: skip_file

import peewee
from peewee import JOIN
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR2,
    Gaia_DR2,
    CatalogFromSDSS_DR19p_Speclite,
    SDSS_DR19p_Speclite,
    Gaia_unWISE_AGN,
)
# DEBUG STUFF TO USE TEMP TABLE
# CatalogToSDSS_DR19p_Speclite._meta.table_name = 'temp_catalog_to_sdss_dr19p_speclite'
# CatalogToSDSS_DR19p_Speclite._meta._schema = 'sandbox'

from target_selection.cartons.base import BaseCarton

# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms
#
# This module provides the following BHM cartons:
# bhm_gua_dark
# bhm_gua_dark_d3
# bhm_gua_bright
#
# Updated by TD on 27/01/2021 to add v0.5 improvements
# switch to a q3c join to sdss_dr16_specobj

"""
  pseudo-SQL

    both Cartons:
        SELECT catalogid,{derived fields}
        FROM catalog AS c
        JOIN catalog_to_tic AS c2tic ON ...
        JOIN tic_v8 AS tic ON ...
        JOIN gaia_dr2_source AS g ON ...
        JOIN gaia_unwise_agn AS t ON g.source_id = t.gaia_sourceid
        LEFT OUTER JOIN sdss_dr16_specobj AS s
          ON ( q3c_join(s.ra,s.dec,c.ra,c.dec,{match_radius_spectro})
               AND s.zwarning = 0
               AND s.snmedianl > 2.0
               AND s.zerr < 0.01
               AND s.scienceprimary > 0
             )
        WHERE
                c.version_id = {version_id}
            AND c2tic.version_id = {version_id}
            AND c2tic.best = True
            AND t.prob_rf > 0.8
            AND s.specobjid = NULL

    # bhm_gua_dark carton:
            AND ( t.g > 16.5 AND
                  t.rp > 16.5 AND
                  (t.g < 21.2 OR t.rp < 21.0 )
                )

    # bhm_gua_bright carton:
            AND ( t.g > 13.0 AND
                  t.rp > 13.5 AND
                  (t.g < 18.5 OR t.rp < 18.5)
                )
"""


class BhmGuaBaseCarton(BaseCarton):
    """
    Parent class that provides the basic selections for both Gaia UnWISE AGN cartons
    To be sub-classed, not to be called directly.

    To get from Catalog to GUA we join tables via :
    Catalog -> CatalogToGaia_DR2 -> Gaia_DR2 -> Gaia_unWISE_AGN

    Cadence+priority per target depends on brightness
    Keep the two bhm_gua_* cartons separate to enable overlap in brightness ranges
    """

    name = "bhm_gua_base"
    category = "science"
    mapper = "BHM"
    program = "bhm_filler"
    tile = False
    priority = None
    cadence = None
    instrument = "BOSS"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2g2 = CatalogToGaia_DR2.alias()
        g2 = Gaia_DR2.alias()
        t = Gaia_unWISE_AGN.alias()

        spec_sn_thresh = self.parameters["spec_sn_thresh"]
        spec_z_err_thresh = self.parameters["spec_z_err_thresh"]

        # #########################################################################
        # prepare the spectroscopy catalogue

        # SDSS DR19p
        # downslect only 'good' spectra
        c2s19 = CatalogFromSDSS_DR19p_Speclite.alias()
        ss19 = SDSS_DR19p_Speclite.alias()
        s19 = (
            ss19.select(
                ss19.pk.alias("s19_pk"),
            )
            .where(
                ss19.sn_median_all >= spec_sn_thresh,
                ss19.zwarning == 0,
                ss19.z_err <= spec_z_err_thresh,
                ss19.z_err > 0.0,
                ss19.specprimary > 0,
            )
            .alias("s19")
        )

        # set the Carton priority+values here - read from yaml
        priority = peewee.Value(int(self.parameters.get("priority", 10000)))
        value = peewee.Value(self.parameters.get("value", 1.0)).cast("float")
        inertial = peewee.Value(True)
        cadence = peewee.Value(self.parameters["cadence"])
        instrument = peewee.Value(self.instrument)

        # compute transformed SDSS mags for pointlike and extended sources separately
        # transform the Gaia dr2 G,BP,RP into sdss psfmag griz

        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ if($3~/sdss_psfmag/){pe="p"} else if ($3~/sdss_fiber2mag/){pe="e"} else{pe="error"}; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  bhm_gua/gdr2_mag_to_sdss_psfmag_?_results.log  # noqa
        coeffs = {
            "g3_p": 0.184158,
            "g2_p": -0.457316,
            "g1_p": 0.553505,
            "g0_p": -0.029152,
            "i3_p": 0.709818,
            "i2_p": -2.207549,
            "i1_p": 1.520957,
            "i0_p": -0.417666,
            "r3_p": 0.241611,
            "r2_p": -0.803702,
            "r1_p": 0.599944,
            "r0_p": -0.119959,
            "z3_p": 0.893988,
            "z2_p": -2.759177,
            "z1_p": 1.651668,
            "z0_p": -0.440676,
        }

        #  bp_rp = t.bp - t.rp
        #  g = (t.g + coeffs['g0_p'] + coeffs['g1_p'] * bp_rp + coeffs['g2_p'] * bp_rp * bp_rp +
        #       coeffs['g3_p'] * bp_rp * bp_rp * bp_rp)
        #  r = (t.g + coeffs['r0_p'] + coeffs['r1_p'] * bp_rp + coeffs['r2_p'] * bp_rp * bp_rp +
        #       coeffs['r3_p'] * bp_rp * bp_rp * bp_rp)
        #  i = (t.g + coeffs['i0_p'] + coeffs['i1_p'] * bp_rp + coeffs['i2_p'] * bp_rp * bp_rp +
        #       coeffs['i3_p'] * bp_rp * bp_rp * bp_rp)
        #  z = (t.g + coeffs['z0_p'] + coeffs['z1_p'] * bp_rp + coeffs['z2_p'] * bp_rp * bp_rp +
        #       coeffs['z3_p'] * bp_rp * bp_rp * bp_rp)

        g = (
            g2.phot_g_mean_mag
            + coeffs["g0_p"]
            + coeffs["g1_p"] * g2.bp_rp
            + coeffs["g2_p"] * g2.bp_rp * g2.bp_rp
            + coeffs["g3_p"] * g2.bp_rp * g2.bp_rp * g2.bp_rp
        )
        r = (
            g2.phot_g_mean_mag
            + coeffs["r0_p"]
            + coeffs["r1_p"] * g2.bp_rp
            + coeffs["r2_p"] * g2.bp_rp * g2.bp_rp
            + coeffs["r3_p"] * g2.bp_rp * g2.bp_rp * g2.bp_rp
        )
        i = (
            g2.phot_g_mean_mag
            + coeffs["i0_p"]
            + coeffs["i1_p"] * g2.bp_rp
            + coeffs["i2_p"] * g2.bp_rp * g2.bp_rp
            + coeffs["i3_p"] * g2.bp_rp * g2.bp_rp * g2.bp_rp
        )
        z = (
            g2.phot_g_mean_mag
            + coeffs["z0_p"]
            + coeffs["z1_p"] * g2.bp_rp
            + coeffs["z2_p"] * g2.bp_rp * g2.bp_rp
            + coeffs["z3_p"] * g2.bp_rp * g2.bp_rp * g2.bp_rp
        )

        # validity checks - set limits semi-manually
        bp_rp_min = 0.0
        bp_rp_max = 1.8
        valid = (
            g2.phot_g_mean_mag.between(0.1, 29.9)
            & g2.phot_bp_mean_mag.between(0.1, 29.9)
            & g2.phot_rp_mean_mag.between(0.1, 29.9)
            & g2.bp_rp.between(bp_rp_min, bp_rp_max)
        )

        opt_prov = peewee.Case(None, ((valid, "sdss_psfmag_from_gaiadr2"),), "undefined")
        magnitude_g = peewee.Case(None, ((valid, g),), "NaN")
        magnitude_r = peewee.Case(None, ((valid, r),), "NaN")
        magnitude_i = peewee.Case(None, ((valid, i),), "NaN")
        magnitude_z = peewee.Case(None, ((valid, z),), "NaN")

        # Create temporary tables for the base query and the Q3C cross-match
        # tables.

        query = (
            c.select(
                c.catalogid,
                c.ra,  # extra
                c.dec,  # extra
                t.gaia_sourceid.alias("gaia_dr2_source_id"),  # extra
                t.unwise_objid,  # extra
                priority.alias("priority"),
                value.alias("value"),
                inertial.alias("inertial"),
                cadence.alias("cadence"),
                instrument.alias("instrument"),
                opt_prov.alias("optical_prov"),
                magnitude_g.alias("g"),
                magnitude_r.alias("r"),
                magnitude_i.alias("i"),
                magnitude_z.alias("z"),
                g2.phot_g_mean_mag.alias("gaia_g"),
                g2.phot_bp_mean_mag.alias("bp"),
                g2.phot_rp_mean_mag.alias("rp"),
                t.w1.alias("gua_w1"),  # extra
                t.w2.alias("gua_w2"),  # extra
                t.prob_rf.alias("gua_prob_rf"),  # extra
                t.phot_z.alias("gua_phot_z"),  # extra
                s19.c.s19_pk.alias("s19_pk"),  # extra
                # rely on the centralised magnitude routines for 'real' griz, bp,rp,gaia_g
            )
            .join(c2g2)
            .join(g2, on=(c2g2.target_id == g2.source_id))
            .join(t, on=(g2.source_id == t.gaia_sourceid))
            # joining to spectroscopy
            .switch(c)
            .join(c2s19, JOIN.LEFT_OUTER)
            .join(s19, JOIN.LEFT_OUTER, on=(s19.c.s19_pk == c2s19.target_id))
            .where(
                c.version_id == version_id,
                c2g2.version_id == version_id,
                fn.coalesce(c2s19.version_id, version_id) == version_id,
                c2g2.best >> True,
                # fn.coalesce(c2s19.best, True) >> True,
            )
            .where(
                (t.prob_rf >= self.parameters["prob_rf_min"]),
                (g2.phot_g_mean_mag >= self.parameters["mag_g_min"]),
                (g2.phot_rp_mean_mag >= self.parameters["mag_rp_min"]),
                (
                    (g2.phot_g_mean_mag < self.parameters["mag_g_max"])
                    | (g2.phot_rp_mean_mag < self.parameters["mag_rp_max"])
                ),
            )
            # then reject any GUA targets with existing good DR19p spectroscopy
            .where(s19.c.s19_pk.is_null(True))
            # avoid duplicates - trust the gaia ids in the GUA parent sample
            .distinct([t.gaia_sourceid])
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    c.ra, c.dec, query_region[0], query_region[1], query_region[2]
                )
            )

        return query


class BhmGuaDarkCarton(BhmGuaBaseCarton):
    """
    -------  bhm_gua_dark   ------
    SQL as above plus:

        AND ( gua.g > 16.x AND gua.rp > 16.x)
    """

    name = "bhm_gua_dark"


class BhmGuaDarkD3Carton(BhmGuaBaseCarton):
    name = "bhm_gua_dark_d3"


class BhmGuaBrightCarton(BhmGuaBaseCarton):
    """
    ------  bhm_gua_bright   ------
    SQL as above plus:

        AND ( gua.g < 18.x OR gua.rp < 18.x)

    """

    name = "bhm_gua_bright"
